from __future__ import annotations

import os
from abc import ABC
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Sequence, TYPE_CHECKING, Union

from common.hash_utils import HashState
from fs_utils import normalize_path
from utils import BuildException

if TYPE_CHECKING:
    from state import Rule

Env = Dict[str, Union[str, List[str]]]


@dataclass
class ContextualRelativePath:
    path: str
    context: Context

    def resolve(self):
        return os.path.relpath(self.path, self.context.cwd)

    def __str__(self):
        return self.resolve()


class DotDict:
    def __init__(self):
        self.items = {}

    def __getitem__(self, item):
        return self.items[item]

    def __getattr__(self, item):
        return self.items[item]

    def __setitem__(self, key, value):
        self.items[key] = value


class Context(ABC):
    def __init__(self, cwd: str, macros: Dict[str, Callable]):
        self.cwd = cwd
        self.deps = DotDict()
        for name, macro in macros.items():
            if hasattr(self, name):
                raise BuildException(f"Macro {name} shadows existing Context attribute")
            setattr(
                self,
                name,
                lambda *args, __macro=macro, **kwargs: __macro(self, *args, **kwargs),
            )

    def absolute(self, path: str):
        path = str(path)  # cast contextual paths back to strings
        return normalize_path(self.cwd, path)

    def relative(self, path: str):
        if isinstance(path, ContextualRelativePath):
            return path
        return ContextualRelativePath(self.absolute(path), self)

    def _resolve(self, dep: str):
        dep = str(dep)
        if dep.startswith(":"):
            return dep
        else:
            return self.absolute(dep)

    def chdir(self, dest: str):
        self.cwd = normalize_path(self.cwd, dest)

    def sh(self, cmd: str, env: Env = None):
        raise NotImplementedError

    def add_dep(self, dep: str, *, load_provided=False, defer=False):
        raise NotImplementedError

    def add_deps(self, deps: Sequence[str], *, load_provided=False, defer=False):
        for dep in deps:
            self.add_dep(dep, load_provided=load_provided, defer=defer)

    def input(self, sh: Optional[str], *, env: Env = None):
        raise NotImplementedError


class MemorizeContext(Context):
    never_defer = False

    def __init__(
        self,
        cwd: str,
        macros: Dict[str, Callable],
        hashstate: HashState,
        dep_fetcher: Callable[[str], "Rule"],
    ):
        super().__init__(cwd, macros)
        self.hashstate = hashstate
        self.dep_fetcher = dep_fetcher
        self.inputs = []
        self.deferred_inputs = []
        self.uses_dynamic_inputs = False
        self.hashstate.record(self.absolute("."))

    def chdir(self, dest: str):
        super().chdir(dest)
        self.hashstate.record("chdir", dest)

    def sh(self, cmd: str, env: Env = None):
        self.hashstate.record("sh", cmd, env)

    def add_dep(self, dep: str, *, load_provided=False, defer=False):
        if load_provided and defer:
            raise BuildException("Cannot load provided value of a deferred dependency")

        dep = str(dep)

        if (
            dep.startswith(":")
            and not defer
            and not load_provided
            and not self.never_defer
        ):
            defer = True

        if defer and self.never_defer:
            raise BuildException("Setup rules cannot defer dependencies")

        if defer:
            self.hashstate.record("add_deferred_dep", dep)
            self.deferred_inputs.append(dep)
        else:
            self.hashstate.record("add_dep", dep)
            self.inputs.append(self._resolve(dep))

        if load_provided:
            rule = self.dep_fetcher(self._resolve(dep))
            if rule is None:
                raise BuildException("Cannot load provided value of a source file")
            self.deps[dep] = rule.provided_value
            if rule.name:
                self.deps[rule.name] = rule.provided_value

        return defer

    def input(self, sh: Optional[str], *, env: Env = None):
        self.uses_dynamic_inputs = True
        self.hashstate.record("input", sh, env)



import hashlib
import os
from os.path import dirname
from pathlib import Path
from typing import List

from fs_utils import copy_helper, hash_file
from state import Rule
from utils import BuildException, CacheMiss

CLOUD_BUCKET_PREFIX = "gs://"
AUX_CACHE = ".aux_cache"

STATS = dict(hits=0, misses=0, inserts=0)


def get_bucket(cache_directory: str):
    if cache_directory.startswith(CLOUD_BUCKET_PREFIX):
        from google.cloud import storage

        client = storage.Client()
        return client.bucket(cache_directory[len(CLOUD_BUCKET_PREFIX) :])


def make_cache_load(cache_directory: str, *, is_aux=False):
    bucket = get_bucket(cache_directory)
    delta = 1 if not is_aux else 0

    def cache_load_string(state: str, target: str) -> str:
        from google.cloud.exceptions import NotFound

        if bucket:
            try:
                out = aux_load_string(state, target)
                STATS["hits"] += delta
                return out
            except CacheMiss:
                dest = str(Path(state).joinpath(target))
                try:
                    out = bucket.blob(dest).download_as_string().decode("utf-8")
                    STATS["hits"] += delta
                except NotFound:
                    STATS["misses"] += delta
                    raise CacheMiss
                else:
                    # cache it on disk
                    aux_store_string(state, target, out)
                    return out
        else:
            dest = Path(cache_directory).joinpath(state).joinpath(target)
            try:
                with open(dest, "r") as f:
                    out = f.read()
                    STATS["hits"] += delta
                    return out
            except FileNotFoundError:
                STATS["misses"] += delta
                raise CacheMiss

    def cache_load_files(cache_key: str, rule: Rule, dest_root: str) -> List[str]:
        cache_location, cache_paths = get_cache_output_paths(
            cache_directory, rule, cache_key
        )
        out = []
        if bucket:
            from google.cloud.exceptions import NotFound

            del cache_location
            if not aux_load_files(cache_key, rule, dest_root):
                try:
                    cache_load_string(cache_key, ".touch")
                except CacheMiss:
                    STATS["misses"] += delta
                    raise
                for src_name, cache_path in zip(rule.outputs, cache_paths):
                    cache_path = str(Path(cache_key).joinpath(cache_path))
                    os.makedirs(dest_root, exist_ok=True)
                    try:
                        if src_name.endswith("/"):
                            os.makedirs(
                                Path(dest_root).joinpath(src_name), exist_ok=True
                            )
                            blobs = list(bucket.list_blobs(prefix=cache_path))
                            for blob in blobs:
                                filename = blob.name[len(cache_path) + 1 :]
                                target = str(
                                    Path(dest_root)
                                    .joinpath(src_name)
                                    .joinpath(filename)
                                )
                                os.makedirs(dirname(target), exist_ok=True)
                                blob.download_to_filename(target)
                                STATS["hits"] += delta
                                out.append(os.path.join(src_name, filename))
                        else:
                            target = os.path.join(dest_root, src_name)
                            os.makedirs(dirname(target), exist_ok=True)
                            bucket.blob(cache_path).download_to_filename(target)
                            out.append(target)
                            STATS["hits"] += delta
                    except NotFound:
                        STATS["misses"] += delta
                        raise CacheMiss
                # now that we have fetched, let's cache it on disk
                aux_store_files(cache_key, rule, dest_root)
            return out
        else:
            if not os.path.exists(cache_location):
                STATS["misses"] += delta
                raise CacheMiss
            try:
                _, out = copy_helper(
                    src_root=cache_location,
                    src_names=cache_paths,
                    dest_root=dest_root,
                    dest_names=rule.outputs,
                )
                STATS["hits"] += delta
            except FileNotFoundError:
                raise BuildException(
                    "Cache corrupted. This should never happen unless you modified the cache "
                    "directory manually! If so, delete the cache directory and try again."
                )
            return out

    return cache_load_string, cache_load_files


def make_cache_store(cache_directory: str, *, is_aux=False):
    bucket = get_bucket(cache_directory)
    delta = 1 if not is_aux else 0

    def cache_store_string(state: str, target: str, data: str):
        if bucket:
            try:
                prev_saved = aux_load_string(state, target)
            except CacheMiss:
                prev_saved = None
            aux_store_string(state, target, data)
            if prev_saved != data:
                STATS["inserts"] += delta
                bucket.blob(str(Path(state).joinpath(target))).upload_from_string(data)
        else:
            cache_target = Path(cache_directory).joinpath(state).joinpath(target)
            os.makedirs(os.path.dirname(cache_target), exist_ok=True)
            STATS["inserts"] += delta
            cache_target.write_text(data)

    def cache_store_files(cache_key: str, rule: Rule, output_root: str) -> List[str]:
        cache_location, cache_paths = get_cache_output_paths(
            cache_directory, rule, cache_key
        )

        cache_store_string(cache_key, ".touch", "")

        out = []

        if bucket:
            del cache_location  # just to be safe

            for src_name, cache_path in zip(rule.outputs, cache_paths):
                if src_name.endswith("/"):
                    dir_root = Path(output_root).joinpath(src_name)
                    for path, subdirs, files in os.walk(dir_root):
                        path = os.path.relpath(path, dir_root)
                        for name in files:
                            #               output_root -> src_name/ -> [path -> name]
                            # <cache_base> -> cache_key -> cache_path -> [path -> name]
                            target = Path(path).joinpath(name)
                            src_loc = (
                                Path(output_root).joinpath(src_name).joinpath(target)
                            )
                            aux_cache_loc = (
                                Path(AUX_CACHE)
                                .joinpath(cache_key)
                                .joinpath(cache_path)
                                .joinpath(target)
                            )
                            if not os.path.exists(aux_cache_loc) or (
                                hash_file(src_loc) != hash_file(aux_cache_loc)
                            ):
                                STATS["inserts"] += delta
                                bucket.blob(
                                    str(
                                        Path(cache_key)
                                        .joinpath(cache_path)
                                        .joinpath(target)
                                    )
                                ).upload_from_filename(str(src_loc))
                            out.append(os.path.join(src_name, target))
                else:
                    #                 output_root -> src_name
                    # <cache_base> -> cache_key -> cache_path

                    target = Path(output_root).joinpath(src_name)
                    aux_cache_loc = (
                        Path(AUX_CACHE).joinpath(cache_key).joinpath(cache_path)
                    )
                    if not os.path.exists(aux_cache_loc) or (
                        hash_file(target) != hash_file(aux_cache_loc)
                    ):
                        STATS["inserts"] += delta
                        bucket.blob(
                            str(Path(cache_key).joinpath(cache_path)),
                        ).upload_from_filename(
                            str(Path(output_root).joinpath(src_name)),
                        )
                    out.append(os.path.join(src_name, target))

            aux_store_files(cache_key, rule, output_root)

        else:
            STATS["inserts"] += delta
            out, _ = copy_helper(
                src_root=output_root,
                src_names=rule.outputs,
                dest_root=cache_location,
                dest_names=cache_paths,
            )
        return out

    return cache_store_string, cache_store_files


def get_cache_output_paths(cache_directory: str, rule: Rule, cache_key: str):
    cache_location = Path(cache_directory).joinpath(cache_key)
    keys = [
        hashlib.md5(output_path.encode("utf-8")).hexdigest()
        for output_path in rule.outputs
    ]
    for i, (output_path, key) in enumerate(zip(rule.outputs, keys)):
        if output_path.endswith("/"):
            keys[i] += "/"
    return cache_location, keys


aux_load_string, aux_load_files = make_cache_load(AUX_CACHE, is_aux=True)
aux_store_string, aux_store_files = make_cache_store(AUX_CACHE, is_aux=True)



import os
import traceback
from typing import Callable, Dict, List, Optional, Sequence, Set

from colorama import Style

from cache import make_cache_load, make_cache_store
from context import Env
from execution import ExecutionContext
from fs_utils import hash_file
from monitoring import create_status_monitor, log
from state import Rule, TargetLookup
from utils import BuildException, CacheMiss, MissingDependency
from common.hash_utils import HashState


class WorkspaceExecutionContext(ExecutionContext):
    never_defer = True

    def __init__(
        self,
        hashstate: HashState,
        dep_fetcher: Callable[[str], Rule],
    ):
        super().__init__(
            os.curdir,
            os.curdir,
            {},
            hashstate,
            dep_fetcher,
            lambda *args: None,
        )

    def normalize(self, env: Optional[Env]):
        if env is None:
            env = {}
        return super().normalize({**env, **os.environ})


def initialize_workspace(
    setup_rule_lookup: TargetLookup,
    setup_targets: List[str],
    state_directory: str,
    quiet: bool,
):
    # we don't need the indirect lookup as we only have rule and source deps
    direct_lookup: Dict[str, Rule] = setup_rule_lookup.direct_lookup
    work_queue = []
    for setup_target in setup_targets:
        if setup_target not in direct_lookup:
            raise BuildException(f"Unknown or unspecified setup target {setup_target}")
        work_queue.append(direct_lookup[setup_target])

    rebuilt: Set[str] = set()
    ready: Set[str] = set()

    cache_load_string, _ = make_cache_load(state_directory)
    cache_store_string, _ = make_cache_store(state_directory)

    if work_queue:
        status_monitor = create_status_monitor(1, quiet)
        status_monitor.move(total=len(work_queue))

    def dep_fetcher(dep):
        if dep.startswith(":"):
            if dep not in direct_lookup:
                raise BuildException(f"Unable to find setup rule {dep}")
            dep_rule = direct_lookup[dep]
            log(f"Looking up setup rule {dep}")
            if dep_rule not in ready:
                raise MissingDependency(dep)
            return dep_rule

    while work_queue:
        todo = work_queue.pop()
        log(f"Popping setup rule {todo} off work queue")
        try:
            if todo.name is None:
                raise BuildException(
                    f"All setup rules must have names, but {todo} does not."
                )

            hashstate = HashState()
            ctx = WorkspaceExecutionContext(hashstate, dep_fetcher)
            unchecked_rules = []

            try:
                todo.set_provided_value(
                    todo.impl(ctx),
                    None,
                    ctx.inputs,
                    ctx.deferred_inputs,
                    [],  # todo: implement output providers for setup rules
                )
                if ctx.out_of_date_deps:
                    raise MissingDependency(*ctx.out_of_date_deps)
            except MissingDependency as e:
                unchecked_rules = [direct_lookup[x] for x in e.paths]

            if unchecked_rules:
                for dep in unchecked_rules:
                    if dep not in work_queue:
                        log(f"Setup rule {todo} is enqueuing {dep}")
                        status_monitor.move(total=1)
                        work_queue.append(dep)
                    else:
                        log(
                            f"Setup rule {todo} is waiting on {dep}, which is already enqueued"
                        )
                    dep.runtime_dependents.add(todo)
                    todo.pending_rule_dependencies.add(dep)
            else:
                log(
                    f"Setup rule {todo} ran with inputs {ctx.inputs + ctx.deferred_inputs}"
                )
                for dep in ctx.inputs + ctx.deferred_inputs:
                    if dep.startswith(":"):
                        continue
                    try:
                        hashstate.record(dep)
                        hashstate.update(hash_file(dep))
                    except FileNotFoundError:
                        raise BuildException(f"Source file {dep} not found.")

                try:
                    ok = cache_load_string("workspace", todo.name) == hashstate.state()
                    if not ok:
                        log(f"State mismatch for rule {todo}, need to rerun")
                except CacheMiss:
                    log(f"State not found for rule {todo}, need to run for first time")
                    ok = False

                for dep in ctx.inputs + ctx.deferred_inputs:
                    if dep.startswith(":"):
                        if direct_lookup[dep] in rebuilt:
                            log(
                                f"Dependency {dep} of setup rule {todo} was rebuilt, so we must rebuild {todo} as well"
                            )
                            ok = False

                for out in todo.outputs:
                    if not os.path.exists(out):
                        log(
                            f"Output {out} is missing for setup rule {todo}, forcing rerun"
                        )
                        ok = False
                        break

                if not ok:
                    # we need to fully run
                    log(f"Fully running setup rule {todo}")
                    ctx.run_shell_queue()
                    rebuilt.add(todo)
                    cache_store_string("workspace", todo.name, hashstate.state())

                # either way, now we can trigger our dependents
                ready.add(todo)
                for dep in todo.runtime_dependents:
                    dep.pending_rule_dependencies.remove(todo)
                    if not dep.pending_rule_dependencies:
                        work_queue.append(dep)
                        status_monitor.move(total=1)

            status_monitor.move(curr=1)
        except Exception as e:
            if not isinstance(e, BuildException):
                suffix = f"\n{Style.RESET_ALL}" + traceback.format_exc()
            else:
                suffix = ""
            status_monitor.stop()
            raise BuildException(
                f"Error while executing rule {todo}: " + str(e) + suffix
            )



class BuildException(Exception):
    pass


# thrown internally by contexts to exit early, should always be caught
class CacheMiss(Exception):
    pass


class MissingDependency(Exception):
    def __init__(self, *paths: str):
        self.paths = paths



from __future__ import annotations

from dataclasses import dataclass
from functools import lru_cache
from typing import Callable, Iterable, List, Sequence, Union

from common.hash_utils import HashState
from fs_utils import hash_file
from utils import BuildException


class ProviderMetaclass(type):
    def __str__(self):
        return "<Provider>"


@dataclass
class AbstractProvider(metaclass=ProviderMetaclass):
    value: object


def provider(*args):
    if len(args) > 1:
        raise BuildException("provider() takes at most one argument")
    if args:

        class Provider(AbstractProvider):
            def __init__(self):
                super().__init__(args[0])

    else:

        class Provider(AbstractProvider):
            pass

    return Provider


@dataclass(frozen=True, eq=False, repr=False)
class DepSet:
    # children are either file paths, or depsets (NOT rules)
    children: Sequence[Union[DepSet, List[str]]]

    def __init__(self, *args: Union[DepSet, str]):
        object.__setattr__(self, "children", list(args))
        if any(
            ".cache" in arg or ".scratch" in arg or arg.endswith("/")
            for arg in args
            if isinstance(arg, str)
        ):
            raise BuildException("?", args)

    @lru_cache
    def _hash(self) -> str:
        print("[WARNING] DepSet hashing is experimental and may corrupt caches")
        hashstate = HashState()
        for child in self.children:
            if isinstance(child, DepSet):
                hashstate.record(child._hash())
            else:
                assert (
                    isinstance(child, str)
                    and not child.endswith("/")
                    and not child.endswith(":")
                ), "Depsets only hold files or other depsets"
                hashstate.record(child)
                hashstate.update(hash_file(child))
        return hashstate.state()

    def to_iter(self):
        explored = set()
        pending = [self]
        while pending:
            todo = pending.pop()
            if todo in explored:
                continue
            explored.add(todo)
            if isinstance(todo, DepSet):
                pending.extend(todo.children)
            else:
                yield f"//{todo}"


class LazyDepSet(DepSet):
    gen_children: Callable[[], Union[DepSet, List[str]]]

    def __init__(self, gen_children: Callable[[], Iterable[Union[DepSet, str]]]):
        assert callable(gen_children)
        object.__setattr__(self, "gen_children", gen_children)

    @property
    def children(self):
        if not hasattr(self, "_children"):
            forced = self.gen_children()
            object.__setattr__(
                self,
                "_children",
                forced if isinstance(forced, DepSet) else list(forced),
            )
        return self._children


class IterableDepSet(DepSet):
    def __iter__(self):
        return self.to_iter()

    def __add__(self, other: Union[DepSet, List[str]]):
        if isinstance(other, DepSet):
            return IterableDepSet(self, other)
        return NotImplemented

    def __bool__(self):
        return any(True for _ in self)


class GlobDepSet(LazyDepSet, IterableDepSet):
    pass


TransitiveDepsProvider = provider()
TransitiveOutputProvider = provider()
DepsProvider = provider()
OutputProvider = provider()



from typing import Callable, Dict

from cache import make_cache_load
from common.hash_utils import HashState
from context import Env, MemorizeContext
from fs_utils import hash_file
from monitoring import log
from state import BuildState, Rule
from utils import CacheMiss, MissingDependency


class PreviewContext(MemorizeContext):
    def __init__(
        self,
        cwd: str,
        macros: Dict[str, Callable],
        hashstate: HashState,
        dep_fetcher: Callable[[str], Rule],
        cache_fetcher: Callable[[str, str], str],
    ):
        super().__init__(cwd, macros, hashstate, dep_fetcher)
        self.cache_fetcher = cache_fetcher

    def input(self, sh: str = None, *, env: Env = None):
        state = self.hashstate.state()
        super().input(sh, env=env)
        return self.cache_fetcher(state, HashState().record(sh, env).state())


def make_dep_fetcher(build_state: BuildState):
    def dep_fetcher(input_path, type: str = "rule"):
        try:
            rule = None
            if input_path not in build_state.source_files:
                rule = build_state.target_rule_lookup.try_lookup(input_path)
                # this input may be stale / unbuilt / no longer exists
                # if so, do not read it, but instead throw MissingDependency
                if rule is None or rule not in build_state.ready:
                    raise MissingDependency(input_path)
                # so it's already ready for use!

            if type == "hash":
                return hash_file(input_path)
            elif type == "rule":
                return rule
            else:
                raise Exception(f"Unknown dep type {type}")
        except FileNotFoundError:
            raise MissingDependency(input_path)

    return dep_fetcher


def get_deps(build_state: BuildState, rule: Rule, *, skip_cache_key: bool):
    """
    Use static dependencies and caches to try and identify as *many*
    needed dependencies as possible, without *any* spurious dependencies.
    """
    hashstate = HashState()
    cache_load_string, _ = make_cache_load(build_state.cache_directory)
    dep_fetcher = make_dep_fetcher(build_state)

    ctx = PreviewContext(
        rule.location,
        build_state.macros,
        hashstate,
        dep_fetcher,
        cache_load_string,
    )

    ok = False
    provided_value = None
    try:
        log(f"Running impl of {rule} to discover dependencies")
        if not skip_cache_key:
            for out in rule.outputs:
                # needed so that if we ask for another output, we don't panic if it's not in the cache
                hashstate.record(out)
        provided_value = rule.impl(ctx)
        log(f"Impl of {rule} completed with deps: {ctx.inputs}")
        ok = True
    except CacheMiss:
        log(f"Cache miss while running impl of {rule}")
        pass  # stops context execution
    except MissingDependency as e:
        log(f"Dependencies {e.paths} were unavailable while running impl of {rule}")
        pass  # dep already added to ctx.inputs
    except Exception as e:
        print(
            "Error occurred during PreviewExecution. This may be normal, if a cached file that has not "
            "yet been reported / processed has been changed. However, it may also be an internal error, so "
            "it is being logged here. If it is an internal error, please contact the maintainer."
        )
        print(repr(e))
        if not ctx.uses_dynamic_inputs:
            raise
    # if `ok`, hash loaded dynamic dependencies
    if ok:
        log(f"Inputs and dependencies resolved for {rule}")
        for input_path in ctx.inputs:
            if input_path.startswith(":"):
                input_dep = build_state.target_rule_lookup.try_lookup(input_path)
                if input_dep is None or input_dep not in build_state.ready:
                    ok = False
                    log(
                        f"Rule dependency {input_path} is not yet ready (or does not exist)"
                    )
                    break
            else:
                if not skip_cache_key:
                    hashstate.update(input_path.encode("utf-8"))
                try:
                    data = dep_fetcher(input_path, "rule" if skip_cache_key else "hash")
                except MissingDependency as e:
                    # this dependency was not needed for deps calculation
                    # but is not verified to be up-to-date
                    ok = False
                    log(
                        f"Dependencies {e.paths} were not needed for the impl, but are not up to date"
                    )
                    break
                else:
                    if not skip_cache_key:
                        hashstate.update(data)
        hashstate.record("done")

    return (
        hashstate.state() if ok else None,
        provided_value,
        ctx.inputs,
        ctx.deferred_inputs,
        ctx.uses_dynamic_inputs,
    )



import os
import time
import traceback
from collections import defaultdict
from pathlib import Path
from queue import Empty, Queue
from shutil import rmtree
from subprocess import CalledProcessError

from cache import make_cache_load, make_cache_store
from colorama import Style
from execution import build
from monitoring import log
from preview_execution import get_deps
from state import BuildState
from utils import BuildException, CacheMiss, MissingDependency
from work_queue import enqueue_deps

TIMINGS = defaultdict(int)


def clear_queue(queue: Queue):
    while not queue.empty():
        try:
            queue.get(False)
        except Empty:
            continue
        queue.task_done()


def worker(build_state: BuildState, index: int):
    scratch_path = Path(f".scratch_{index}")
    if scratch_path.exists():
        rmtree(scratch_path, ignore_errors=True)

    _, cache_store_files = make_cache_store(build_state.cache_directory)
    _, cache_load_files = make_cache_load(build_state.cache_directory)

    while True:
        if build_state.failure is not None:
            # every thread needs to clear the queue since otherwise some other thread might still be filling it up
            clear_queue(build_state.work_queue)
            return  # some thread has failed, emergency stop
        todo = build_state.work_queue.get()
        if todo is None:
            return

        start_time = time.time()

        try:
            build_state.status_monitor.update(index, "Parsing: " + str(todo))

            log(f"Target {todo} popped from queue by worker {index}")

            # only from caches, will never run a subprocess
            (
                cache_key,
                provided_value,
                deps,
                deferred_deps,
                uses_dynamic_deps,
            ) = get_deps(build_state, todo, skip_cache_key=todo.do_not_cache)

            if uses_dynamic_deps:
                log("Target", todo, "Uses dynamic deps")

            if cache_key is None:
                # unable to compute cache_key, potentially because not all deps are ready
                log(
                    f"Target {todo} either has unbuilt dependencies, "
                    f"or does not have a cached dynamic dependency resolved"
                )
                deps_ready = not enqueue_deps(
                    build_state,
                    todo,
                    deps,
                    catch_failure=uses_dynamic_deps,
                )
                if deps_ready:
                    log(f"Apparently {todo} is missing an input cache in the impl")
                else:
                    log(f"Apparently {todo} is waiting on unbuilt dependencies")
            else:
                log(f"All the dependencies of target {todo} are ready: {deps}")
                # if the cache_key is ready, *all* the deps must be ready, not just the discoverable deps!
                # unless the cache_key is not actually cached, in which case our inputs() could be wrong, so
                # we have to run in the working directory to verify
                deps_ready = True

            log(f"Enqueuing deferred dependencies for {todo}: {deferred_deps}")
            enqueue_deps(
                build_state, None, deferred_deps, catch_failure=uses_dynamic_deps
            )

            if deps_ready:
                done = False
                # we need to clear any output folders to prevent cache corruption
                # imagine the output folder X/ has a file A, and we generate just X/B, but then cache X/A *and* X/B
                for out in todo.outputs:
                    if out.endswith("/") and Path(out).exists():
                        rmtree(out)

                # check if we're already cached!
                if cache_key and not todo.do_not_cache:
                    try:
                        outputs = cache_load_files(cache_key, todo, os.curdir)
                        log(f"Target {todo} was loaded from the cache")
                        done = True
                    except CacheMiss:
                        pass

                if not done:
                    # time to execute! but *not* inside the lock
                    # when we release the lock, stuff may change outside, but
                    # we don't care since *our* dependencies (so far) are all available
                    log(f"Target {todo} is not in the cache, rerunning...")
                    build_state.status_monitor.update(index, "Building: " + str(todo))
                    # if cache_key is None or we use dynamic dependencies, we haven't finished evaluating
                    # the impl, so we don't know all the dependencies it could need. Therefore, we must
                    # run it in the working directory, so the impl can find the dependencies it needs
                    # Then, we run it *again*, to verify that the dependencies are accurate.
                    # The cache_key could be None due to race conditions, even if we don't use dynamic deps.
                    # Imagine:
                    # - Rule A depends on a provider of rule B, which is not yet built
                    #   (and on C, because of B's provider)
                    # - We determine that rule B is not built, so A depends on B but can't complete impl,
                    #   so A.cache_key = None
                    # - B is built
                    # - We try to enqueue the deps of A. Enqueue_deps responds that B is already built,
                    #   so A.deps_ready=True
                    # - So at this stage, A.cache_key = None, A.uses_dynamic_deps=False, but A.deps is missing C!
                    if uses_dynamic_deps or cache_key is None:
                        log(
                            f"We don't know the dependencies of {todo}, "
                            f"so we are running the impl in the root directory to find out!"
                        )

                        try:
                            _, cache_key = build(
                                build_state,
                                todo,
                                scratch_path=None,
                                skip_cache_key=todo.do_not_cache,
                            )
                        except MissingDependency as d:
                            log(
                                f"Target {todo} failed to fully build because of the missing dynamic "
                                f"dependencies: {d.paths}, requeuing"
                            )
                            scheduled = enqueue_deps(build_state, todo, d.paths)
                            if not scheduled:
                                # due to race conditions, our dependencies are actually all ready now!
                                # no one else will enqueue us, so it is safe to enqueue ourselves
                                build_state.work_queue.put(todo)
                                build_state.status_monitor.move(total=1)
                        else:
                            # now, if no exception has thrown, all the deps are available to the deps finder
                            (
                                alt_cache_key,
                                provided_value,
                                deps,
                                deferred_deps,
                                uses_dynamic_deps,
                            ) = get_deps(
                                build_state, todo, skip_cache_key=todo.do_not_cache
                            )
                            try:
                                provided_value, alt_cache_key_2 = build(
                                    build_state,
                                    todo,
                                    scratch_path=scratch_path,
                                    precomputed_deps=deps,
                                    skip_cache_key=todo.do_not_cache,
                                )
                            except CalledProcessError as e:
                                build_state.status_monitor.stop()
                                print(e.cmd)
                                print(e)
                                raise BuildException(
                                    f"The dependencies for target {todo} are not fully specified, "
                                    f"as it failed to build when provided only with them."
                                )
                            if not todo.do_not_cache:
                                assert (
                                    cache_key == alt_cache_key == alt_cache_key_2
                                ), "An internal error has occurred"
                            done = True
                    else:
                        log(
                            f"We know all the dependencies of {todo}, so we can run it in a sandbox"
                        )
                        provided_value, alt_cache_key = build(
                            build_state,
                            todo,
                            scratch_path=scratch_path,
                            precomputed_deps=deps,
                            skip_cache_key=todo.do_not_cache,
                        )
                        if not todo.do_not_cache:
                            assert (
                                cache_key == alt_cache_key
                            ), "An internal error has occurred"
                        done = True

                    if done:
                        log(f"Target {todo} has been built fully!")
                        if not todo.do_not_cache:
                            outputs = cache_store_files(cache_key, todo, os.curdir)
                        else:
                            outputs = []
                            for out in todo.outputs:
                                if out.endswith("/"):
                                    outputs.extend(
                                        os.path.join(path, filename)
                                        for path, subdirs, files in os.walk(out)
                                        for filename in files
                                    )
                                else:
                                    outputs.append(out)

                        if scratch_path.exists():
                            rmtree(scratch_path, ignore_errors=True)

                if done:
                    waiting_for_deferred = enqueue_deps(
                        build_state, todo, deferred_deps
                    )
                    # We have to wait for deferred dependencies to build because the TransitiveOutputProvider
                    # needs them to complete. However, we can run the build itself before the deferred deps finish.
                    if waiting_for_deferred:
                        log(
                            f"Target {todo} has been built, but is waiting for deferred dependencies"
                        )
                    else:
                        todo.set_provided_value(
                            provided_value, build_state, deps, deferred_deps, outputs
                        )
                        with build_state.scheduling_lock:
                            build_state.ready.add(todo)
                            # no one will ever add us back, since we are in `ready`
                            build_state.scheduled_but_not_ready.remove(todo)
                            # now it's time to set up our dependents
                            # we need to be inside the lock even if we have no dependents, in case
                            # we *gain* dependents from another thread which could have held the lock!
                            for dependent in todo.runtime_dependents:
                                dependent.pending_rule_dependencies.remove(todo)
                                if not dependent.pending_rule_dependencies:
                                    # this guy is ready to go
                                    build_state.work_queue.put(dependent)
                                    build_state.status_monitor.move(total=1)

            # either way, we're done with this task for now
            build_state.status_monitor.move(curr=1)
            build_state.work_queue.task_done()

            # record timing data
            run_time = time.time() - start_time
            TIMINGS[str(todo)] += run_time
        except Exception as e:
            if isinstance(e, BuildException):
                suffix = f"\n{Style.RESET_ALL}"
            else:
                suffix = f"\n{Style.RESET_ALL}" + traceback.format_exc()
            build_state.status_monitor.stop()
            build_state.failure = BuildException(
                f"Error while executing rule {todo}: " + str(e) + suffix
            )
            build_state.work_queue.task_done()



from threading import Thread
from typing import List

from build_worker import worker
from monitoring import create_status_monitor
from state import BuildState
from utils import BuildException


def run_build(
    build_state: BuildState, targets: List[str], num_threads: int, quiet: bool
):
    build_state.status_monitor = create_status_monitor(num_threads, quiet)

    for target in targets:
        root_rule = build_state.target_rule_lookup.try_lookup(
            target
        ) or build_state.target_rule_lookup.lookup(build_state, ":" + target)
        build_state.scheduled_but_not_ready.add(root_rule)
        build_state.work_queue.put(root_rule)
        build_state.status_monitor.move(total=1)

    thread_instances = []
    for i in range(num_threads):
        thread = Thread(target=worker, args=(build_state, i), daemon=True)
        thread_instances.append(thread)
        thread.start()
    build_state.work_queue.join()

    if build_state.failure is not None:
        raise build_state.failure

    for _ in range(num_threads):
        build_state.work_queue.put(None)
    for thread in thread_instances:
        thread.join()

    build_state.status_monitor.stop()

    if build_state.scheduled_but_not_ready:
        # there is a dependency cycle somewhere!
        for root_rule in targets:
            if root_rule in build_state.scheduled_but_not_ready:
                break
        else:
            raise BuildException("An internal error occurred.")
        chain = []
        pos = root_rule
        while True:
            if pos in chain:
                chain.append(pos)
                raise BuildException(
                    f"Circular dependency detected: Rule {pos} depends on itself "
                    f"through the path: {' -> '.join(map(str, chain))}"
                )
            else:
                chain.append(pos)
                pos = next(iter(pos.pending_rule_dependencies))



from __future__ import annotations

import os
import traceback
from json import loads
from shutil import rmtree
from typing import List, Tuple

import click
from colorama import Fore, Style

from build_coordinator import run_build
from build_worker import TIMINGS as BUILD_TIMINGS
from cache import STATS
from common.cli_utils import pretty_print
from fs_utils import find_root, get_repo_files
from loader import TIMINGS as LOAD_TIMINGS, config, load_rules
from monitoring import enable_logging, enable_profiling
from state import BuildState
from utils import BuildException
from workspace_setup import initialize_workspace


def display_error(error: BuildException):
    print(Fore.RED)
    pretty_print("🚫", "Build failed.")
    print(Style.BRIGHT)
    print(error)


@click.command()
@click.argument("targets", metavar="target", required=False, nargs=-1)
@click.option(
    "--profile",
    "-p",
    default=False,
    is_flag=True,
    help="Show performance profiling data, cache hit rates, and log all executed commands.",
)
@click.option(
    "--shell-log",
    "-s",
    default=False,
    is_flag=True,
    help="Log all executed commands.",
)
@click.option(
    "--locate",
    "-l",
    default=False,
    is_flag=True,
    help="Rather than building, locate the BUILD file declaring a particular target.",
)
@click.option(
    "--verbose",
    "-v",
    default=False,
    is_flag=True,
    help="Show detailed logs to identify problems. For most purposes, --profile is a better option.",
)
@click.option(
    "--quiet",
    "-q",
    default=False,
    is_flag=True,
    help="Disable the progress bars normally shown during a build. Does not affect the output of "
    "--locate or --profile, and errors will still be printed if the build fails.",
)
@click.option(
    "--skip-version-check",
    default=False,
    is_flag=True,
    help="Ignore any minimum version requirements specified in the WORKSPACE file.",
)
@click.option(
    "--skip-setup",
    default=False,
    is_flag=True,
    help="Do not install / verify dependencies in the WORKSPACE file.",
)
@click.option(
    "--skip-build",
    default=False,
    is_flag=True,
    help="Do not build the targets passed in. Automatically set when running --locate.",
)
@click.option(
    "--clean",
    "-c",
    default=False,
    is_flag=True,
    help="Clean the output directory specified in the WORKSPACE file so only built targets will be kept.",
)
@click.option(
    "--threads",
    "-t",
    "num_threads",
    default=8,
    help="The number of worker threads used in execution. Increase this number if using a remote build cache.",
)
@click.option(
    "--state-directory",
    default=".state",
    help="The local directory used to keep track of the state of the WORKSPACE install.",
)
@click.option(
    "--cache-directory",
    default=os.getenv("BUILDTOOL_CACHE_DIRECTORY", ".cache"),
    help="The directory used to cache build outputs and intermediate layers. If a cloud storage bucket is passed "
    "in as gs://bucket-name, it will be used alongside an .aux_cache local directory as a source for cached outputs. "
    "Override with the BUILDTOOL_CACHE_DIRECTORY environment variable.",
)
@click.option(
    "--flag",
    "-f",
    "flags",
    type=str,
    multiple=True,
    help="These flags are exposed at load and build time, via `from buildtool import flags`. "
    "Values can be supplied as JSON like --flag KEY=JSON_VALUE, and flags without values default to `true`.",
)
def cli(
    targets: Tuple[str],
    profile: bool,
    shell_log: bool,
    locate: bool,
    verbose: bool,
    quiet: bool,
    skip_version_check: bool,
    skip_setup: bool,
    skip_build: bool,
    clean: bool,
    num_threads: int,
    state_directory: str,
    cache_directory: str,
    flags: List[str],
):
    """
    This is a `make` alternative with a simpler syntax and some useful features.
    """
    try:
        repo_root = find_root()
        os.chdir(repo_root)

        if verbose:
            enable_logging()

        if profile or shell_log:
            enable_profiling()

        flags = [flag.split("=", 1) + ["true"] for flag in flags]
        flags = {flag[0].lower(): loads(flag[1]) for flag in flags}

        if not skip_setup:
            setup_rule_lookup, _ = load_rules(
                flags, workspace=True, skip_version_check=skip_version_check
            )

            setup_targets = [
                target[5:] for target in targets if target.startswith("setup:")
            ]

            if locate and setup_targets:
                raise BuildException(
                    "--locate cannot be used with setup rules - they are declared in WORKSPACE"
                )

            setup_targets = setup_targets or (
                [config.default_setup_rule] if config.default_setup_rule else []
            )

            initialize_workspace(
                setup_rule_lookup,
                setup_targets,
                state_directory,
                quiet,
            )

        target_rule_lookup, macros = load_rules(
            flags, skip_version_check=skip_version_check
        )
        target_rule_lookup.verify()

        all_files = get_repo_files()
        source_files = target_rule_lookup.find_source_files(all_files)

        if profile:
            print("Slow Build / Rules Files (Loading Phase):")
            slowest = sorted(LOAD_TIMINGS, key=lambda x: LOAD_TIMINGS[x], reverse=True)[
                :20
            ]
            for key in slowest:
                print(key, LOAD_TIMINGS[key])

            print()

        need_target = locate or not skip_build

        if not targets and need_target:
            if config.default_build_rule is None:
                raise BuildException("No target provided, and no default target set.")
            targets = [config.default_build_rule]

        if locate:
            for target in targets:
                if target in source_files:
                    raise BuildException(
                        f"Target {target} is a source file, not a build target."
                    )
                rule = target_rule_lookup.try_lookup(target)
                if rule is None and not target.startswith(":"):
                    rule = target_rule_lookup.try_lookup(f":{target}")
                if rule is None:
                    raise BuildException(f"Target {target} was not found.")
                print(
                    f"Target {target} is declared by {rule} in {rule.location}/BUILD."
                )
            exit(0)

        if not skip_build:
            for _ in range(2 if clean else 1):
                if clean:
                    for out_dir in config.output_directories:
                        rmtree(out_dir, ignore_errors=True)
                    for rule in set(target_rule_lookup.direct_lookup.values()) | set(
                        target_rule_lookup.direct_lookup.values()
                    ):
                        rule.pending_rule_dependencies = set()
                        rule.runtime_dependents = set()
                run_build(
                    BuildState(
                        target_rule_lookup=target_rule_lookup,
                        source_files=source_files,
                        cache_directory=cache_directory,
                        macros=macros,
                    ),
                    [target for target in targets if not target.startswith("setup:")],
                    num_threads,
                    quiet,
                )

        if profile:
            print("Slow Rules (Execution Phase):")
            slowest = sorted(
                BUILD_TIMINGS, key=lambda x: BUILD_TIMINGS[x], reverse=True
            )[:20]
            for key in slowest:
                print(key, BUILD_TIMINGS[key])
            print("Cache Statistics")
            print(
                f"{STATS['hits']} cache hits, {STATS['misses']} cache misses, {STATS['inserts']} cache inserts (approx)"
            )

    except BuildException as e:
        display_error(e)
        exit(1)
    except Exception as e:
        display_error(BuildException("Internal error: " + repr(e)))
        print(f"\n{Style.RESET_ALL}" + traceback.format_exc())
        exit(1)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import os
import time
import traceback
from dataclasses import dataclass, field
from glob import glob
from importlib.metadata import version
from typing import Callable, Dict, List, Optional, Sequence, Set, Union

from colorama import Style
from packaging.version import parse

from fs_utils import get_repo_files, normalize_path
from providers import (
    DepSet,
    DepsProvider,
    GlobDepSet,
    OutputProvider,
    TransitiveDepsProvider,
    TransitiveOutputProvider,
    provider,
)
from state import Rule, TargetLookup
from utils import BuildException

LOAD_FRAME_CACHE: Dict[str, Struct] = {}

TIMINGS = {}

start_time_stack = []


@dataclass
class Config:
    # CLI config
    skip_version_check: bool = False

    # loaded config
    default_setup_rule: Optional[str] = None
    default_build_rule: Optional[str] = None
    active: bool = False
    output_directories: List[str] = field(default_factory=list)

    def _check_active(self):
        if not self.active:
            raise BuildException("Cannot use config in this context.")

    @staticmethod
    def _check_rule(rule: str):
        if not rule.startswith(":"):
            raise BuildException(f"Can only register a rule, not {rule}")

    def register_default_setup_rule(self, rule: str):
        self._check_active()
        self._check_rule(rule)
        if self.default_setup_rule is not None:
            raise BuildException(
                f"Default setup rule is already set to {self.default_setup_rule}"
            )
        self.default_setup_rule = rule

    def register_default_build_rule(self, rule: str):
        self._check_active()
        self._check_rule(rule)
        if self.default_build_rule is not None:
            raise BuildException(
                f"Default build rule is already set to {self.default_build_rule}"
            )
        self.default_build_rule = rule

    def register_output_directory(self, path: str):
        self._check_active()
        self.output_directories.append(normalize_path(os.curdir, path))

    def require_buildtool_version(self, min_version: str):
        if self.skip_version_check:
            return
        curr_version = version("buildtool").replace("-", "9999")
        if parse(curr_version) < parse(min_version):
            raise BuildException(
                f"Current buildtool version {curr_version} < {min_version}, the minimum required "
                "for this project. Please upgrade, or pass in --skip-version-check to skip this check."
            )


config = Config()


def make_callback(
    build_root: Optional[str],
    src_files: Set[str],
    # output parameters
    target_rule_lookup: TargetLookup,
    macros: Dict[str, Callable],
):
    make_callback.build_root = build_root
    make_callback.find_cache = {}
    make_callback.target_rule_lookup = target_rule_lookup
    make_callback.macros = macros

    def fail(target):
        raise BuildException(
            f"The target `{target}` is built by multiple rules. Targets can only be produced by a single rule."
        )

    def add_target_rule(target, rule):
        target_rule_lookup = make_callback.target_rule_lookup
        if target.endswith("/"):
            # it's a folder dependency
            if target in target_rule_lookup.location_lookup:
                fail(target)
            target_rule_lookup.location_lookup[target] = rule
        else:
            if target in target_rule_lookup.direct_lookup:
                fail(target)
            target_rule_lookup.direct_lookup[target] = rule

    def callback(
        *,
        name: Optional[str] = None,
        deps: Sequence[str] = (),
        impl: Callable = lambda _: None,
        out: Union[str, Sequence[str]] = (),
        do_not_symlink: bool = False,
        do_not_cache: bool = False,
    ):
        build_root = make_callback.build_root

        if build_root is None:
            raise BuildException(
                "Rules files can only define functions, not invoke callback()"
            )

        if isinstance(out, str):
            out = [out]

        def wrapped_impl(ctx):
            ctx.add_deps(deps)
            return impl(ctx)

        rule = Rule(
            name=name,
            location=build_root,
            impl=wrapped_impl,
            outputs=[normalize_path(build_root, output) for output in out],
            do_not_symlink=do_not_symlink,
            do_not_cache=do_not_cache,
        )
        for output in rule.outputs:
            add_target_rule(output, rule)

        if name is not None:
            add_target_rule(":" + name, rule)

        return f":{name}"

    def find(path, *, unsafe_ignore_extension=False):
        build_root = make_callback.build_root

        if not unsafe_ignore_extension:
            ext = path.split(".")[-1]
            if "/" in ext:
                raise BuildException(
                    "Cannot find() files without specifying an extension"
                )

        if build_root is None and not path.startswith("//"):
            raise BuildException(
                "find() cannot be invoked outside a BUILD file except using a repo-relative path."
            )

        target = normalize_path(build_root, path)

        if target in make_callback.find_cache:
            return make_callback.find_cache[target]

        def gen():
            return sorted(
                os.path.relpath(f, os.curdir)
                for f in glob(
                    target,
                    recursive=True,
                )
                if os.path.relpath(os.path.realpath(f), os.curdir) in src_files
            )

        out = make_callback.find_cache[target] = GlobDepSet(gen)
        return out

    def resolve(path):
        build_root = make_callback.build_root

        if build_root is None:
            raise BuildException(
                "Rules files can only define functions, not invoke resolve(). "
                "If you are in an impl() function, use ctx.relative() instead."
            )

        return "//" + normalize_path(build_root, path)

    def macro(func):
        macros = make_callback.macros
        name = func.__name__
        if name in macros:
            raise BuildException(f"A macro is already defined with name {name}")
        macros[name] = func

    def make_depset(*args):
        return GlobDepSet(
            lambda build_root=make_callback.build_root: (
                dep if isinstance(dep, DepSet) else normalize_path(build_root, dep)
                for dep in (arg() if callable(arg) else arg for arg in args)
            )
        )

    return callback, find, resolve, macro, make_depset


class Struct:
    def __init__(self, entries, default=False):
        self.__dict__.update(entries)
        self.default = default

    def __getattr__(self, item):
        if self.default:
            return None
        else:
            return getattr(super(), item)


def make_load_rules(rules_root: str):
    make_load_rules.rules_root = os.path.dirname(rules_root)

    def load_rules(path):
        path = normalize_path(make_load_rules.rules_root, path)
        if not path.endswith(".py"):
            raise BuildException(f"Cannot import from a non .py file: {path}")

        if path in LOAD_FRAME_CACHE:
            return LOAD_FRAME_CACHE[path]

        start_time_stack.append(time.time())

        old_rules_root = make_load_rules.rules_root
        __builtins__["load"] = make_load_rules(path)

        # We hide the callback here, since you should not be running the
        # callback (or anything else!) in an import, but just providing defs
        frame = {"__builtins__": __builtins__}
        cached_root = make_callback.build_root
        make_callback.build_root = None
        with open(path) as f:
            try:
                exec(compile(f.read(), path, "exec"), frame)
            except Exception:
                raise BuildException(
                    f"Error while processing rules file {path}:\n"
                    + f"\n{Style.RESET_ALL}"
                    + traceback.format_exc()
                )
        make_callback.build_root = cached_root
        make_load_rules.rules_root = old_rules_root

        TIMINGS[path] = load_time = time.time() - start_time_stack.pop()
        start_time_stack[0] += load_time

        out = LOAD_FRAME_CACHE[path] = Struct(frame)
        return out

    return load_rules


def load_rules(
    flags: Dict[str, object],
    *,
    skip_version_check: bool,
    workspace: bool = False,
):
    flags = Struct(flags, default=True)
    src_files = get_repo_files()
    build_files = (
        ["WORKSPACE"]
        if workspace
        else [file for file in src_files if file.split("/")[-1] == "BUILD"]
    )
    target_rule_lookup = TargetLookup()
    macros = {}
    LOAD_FRAME_CACHE.clear()
    callback, find, resolve, macro, make_depset = make_callback(
        None,
        set(src_files),
        target_rule_lookup,
        macros,
    )
    config.skip_version_check = skip_version_check
    for build_file in build_files:
        make_callback.build_root = os.path.dirname(build_file)

        with open(build_file) as f:
            frame = {}
            load = make_load_rules(build_file)

            __builtins__["callback"] = __builtins__["rule"] = callback
            __builtins__["find"] = find
            __builtins__["resolve"] = resolve
            __builtins__["load"] = load
            __builtins__["flags"] = flags
            __builtins__["macro"] = macro
            __builtins__["provider"] = provider
            __builtins__["depset"] = make_depset
            __builtins__["DepsProvider"] = DepsProvider
            __builtins__["OutputProvider"] = OutputProvider
            __builtins__["TransitiveDepsProvider"] = TransitiveDepsProvider
            __builtins__["TransitiveOutputProvider"] = TransitiveOutputProvider

            frame = {
                **frame,
                "__builtins__": __builtins__,
            }

            if workspace:
                __builtins__["config"] = config
                config.active = True

            start_time_stack.append(time.time())
            try:
                exec(compile(f.read(), build_file, "exec"), frame)
                config.active = False
            except Exception:
                raise BuildException(
                    f"Error while processing BUILD file {build_file}:\n"
                    + f"\n{Style.RESET_ALL}"
                    + traceback.format_exc()
                )
            TIMINGS[build_file] = time.time() - start_time_stack.pop()

        make_callback.build_root = None

    return target_rule_lookup, macros



from __future__ import annotations

from collections.abc import Collection
from typing import Optional

from state import BuildState, Rule
from monitoring import log


def enqueue_deps(
    build_state: BuildState,
    rule: Optional[Rule],
    candidate_deps: Collection[str],
    *,
    catch_failure: bool = False,
) -> bool:
    waiting_for_deps = False

    with build_state.scheduling_lock:
        for dep in candidate_deps:
            if dep in build_state.source_files:
                # nothing to do
                continue

            if catch_failure:
                runtime_dep: Rule = build_state.target_rule_lookup.try_lookup(dep)
                if runtime_dep is None:
                    continue
            else:
                runtime_dep: Rule = build_state.target_rule_lookup.lookup(
                    build_state, dep
                )

            if runtime_dep not in build_state.ready:
                waiting_for_deps = True
                if runtime_dep not in build_state.scheduled_but_not_ready:
                    # enqueue dependency
                    log(f"Enqueueing dependency {runtime_dep}")
                    build_state.scheduled_but_not_ready.add(runtime_dep)
                    build_state.work_queue.put(runtime_dep)
                    build_state.status_monitor.move(total=1)
                else:
                    log(f"Waiting on already queued dependency {runtime_dep}")
                if rule:
                    # register task in the already queued / executing dependency
                    # so when it finishes we may be triggered
                    # if rule is None, these are deferred dependencies, so no need to register anything
                    log(f"Registering {rule} to wait on {runtime_dep}")
                    runtime_dep.runtime_dependents.add(rule)
                    rule.pending_rule_dependencies.add(runtime_dep)
    return waiting_for_deps



from __future__ import annotations

import os
import traceback
from pathlib import Path
from subprocess import CalledProcessError
from typing import Callable, Dict, List, Optional

from colorama import Style

from cache import make_cache_store
from common.hash_utils import HashState
from common.shell_utils import sh as run_shell
from context import Env, MemorizeContext
from fs_utils import copy_helper, hash_file
from monitoring import log
from state import BuildState, Rule
from utils import BuildException, MissingDependency


class ExecutionContext(MemorizeContext):
    def __init__(
        self,
        base: str,
        cwd: str,
        macros: Dict[str, Callable],
        hashstate: HashState,
        dep_fetcher: Callable[[str], Rule],
        memorize: Callable[[str, str, str], None],
    ):
        super().__init__(cwd, macros, hashstate, dep_fetcher)
        self.base = base
        self.memorize = memorize
        self.sh_queue = []
        self.out_of_date_deps = []
        os.makedirs(os.path.join(self.base, self.cwd), exist_ok=True)

    def normalize_single(self, path: str):
        if path.startswith("@//"):
            return os.path.abspath(os.path.join(os.curdir, path[3:]))
        elif path.startswith("//"):
            return os.path.abspath(os.path.join(self.base, path[2:]))
        else:
            return path

    def normalize(self, env: Optional[Env]):
        if env is None:
            return {}
        out = {}
        for key in env:
            if isinstance(env[key], str):
                out[key] = self.normalize_single(env[key])
            else:
                out[key] = ":".join(
                    self.normalize_single(component) for component in env[key]
                )
        return out

    def sh(self, cmd: str, env: Env = None):
        super().sh(cmd=cmd, env=env)
        self.sh_queue.append([cmd, self.cwd, self.normalize(env)])

    def run_shell_queue(self):
        for cmd, cwd, env in self.sh_queue:
            log("RUNNING cmd:", cmd)
            run_shell(
                cmd,
                shell=True,
                cwd=os.path.join(self.base, cwd),
                quiet=True,
                capture_output=True,
                inherit_env=False,
                env=env,
            )
        self.sh_queue = []

    def add_dep(self, dep: str, *, load_provided=False, defer=False):
        defer = super().add_dep(dep, load_provided=load_provided, defer=defer)
        if not defer:
            dep = self._resolve(dep)
            try:
                self.dep_fetcher(dep)
            except MissingDependency:
                self.out_of_date_deps.append(dep)

    def input(self, sh: Optional[str] = None, *, env: Env = None):
        # we want the state *before* running the action
        state = self.hashstate.state()
        super().input(sh, env=env)
        if self.out_of_date_deps:
            raise MissingDependency(*self.out_of_date_deps)
        self.run_shell_queue()
        log("RUNNING", sh)
        out = run_shell(
            sh,
            shell=True,
            cwd=os.path.join(self.base, self.cwd),
            capture_output=True,
            quiet=True,
            inherit_env=False,
            env=self.normalize(env),
        ).decode("utf-8")
        self.memorize(state, HashState().record(sh, env).state(), out)
        return out


def build(
    build_state: BuildState,
    rule: Rule,
    *,
    precomputed_deps: Optional[List[str]] = None,
    scratch_path: Optional[Path],
    skip_cache_key: bool,
):
    """
    All the dependencies that can be determined from caches have been
    obtained. Now we need to run. Either we will successfully finish everything,
    or we will get a missing dependency and have to requeue
    """
    cache_store_string, _ = make_cache_store(build_state.cache_directory)

    in_sandbox = scratch_path is not None

    loaded_deps = set()

    def dep_fetcher(dep, *, initial_load=False):
        if dep not in loaded_deps and in_sandbox:
            if not initial_load:
                raise BuildException(
                    f"New dep {dep} found when rerunning rule, it's likely not deterministic!"
                )
            if not dep.startswith(":"):
                log(f"Loading dependency {dep} into sandbox")
                copy_helper(
                    src_root=os.curdir,
                    dest_root=scratch_path,
                    src_names=[dep],
                    symlink=not rule.do_not_symlink,
                )

        # check that these deps are built! Since they may not have been checked by the PreviewExecution.
        dep_rule = None
        if dep not in build_state.source_files:
            dep_rule = build_state.target_rule_lookup.lookup(build_state, dep)
            if dep_rule not in build_state.ready:
                raise MissingDependency(dep)

        loaded_deps.add(dep)

        return dep_rule

    if precomputed_deps:
        assert in_sandbox
        for dep in precomputed_deps:
            dep_fetcher(dep, initial_load=True)

    hashstate = HashState()

    ctx = ExecutionContext(
        scratch_path if in_sandbox else os.curdir,
        rule.location,
        build_state.macros,
        hashstate,
        dep_fetcher,
        cache_store_string,
    )

    try:
        if not skip_cache_key:
            for out in rule.outputs:
                # needed so that if we ask for another output, we don't panic if it's not in the cache
                hashstate.record(out)
        provided_value = rule.impl(ctx)
        if ctx.out_of_date_deps:
            raise MissingDependency(*ctx.out_of_date_deps)
        if in_sandbox:
            ctx.run_shell_queue()
    except CalledProcessError as e:
        raise BuildException(
            "".join(
                [
                    str(e) + "\n",
                    Style.RESET_ALL,
                    f"Location: {scratch_path}\n",
                    f"Working Directory: {scratch_path}/{ctx.cwd}\n",
                    e.stdout.decode("utf-8"),
                    e.stderr.decode("utf-8"),
                ]
            )
        )

    if in_sandbox:
        try:
            copy_helper(
                src_root=scratch_path,
                src_names=rule.outputs,
                dest_root=os.curdir,
            )
        except FileNotFoundError as e:
            raise BuildException(
                f"Output file {e.filename} from rule {rule} was not generated."
            )

    if not skip_cache_key:
        for input_path in ctx.inputs:
            if input_path.startswith(":"):
                # don't hash rule deps
                continue
            hashstate.update(input_path.encode("utf-8"))
            hashstate.update(hash_file(input_path))
    hashstate.record("done")

    return provided_value, hashstate.state() if not skip_cache_key else None
